// Fast SRTP AES encrypt/decrypt.
// Written by Nils Liaaen Corneliusen 2016.
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
// Please refer to the article at https://www.ignorantus.com/pages/srtp_aes_revisited/ for more information.

#include <stdio.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <assert.h>
#include <openssl/aes.h>

#include "srtp_crypt.h"

#define MAX_LEN (65536*16)

static uint64_t clockish( void )
{
    struct timeval tv;
    gettimeofday( &tv, NULL );
    return (((uint64_t)tv.tv_sec)*1000000)+((uint64_t)tv.tv_usec);
}

static void hexdump( char *name, void *src, int len )
{
    uint8_t *in = (uint8_t *)src;
    printf( "%s: ", name );
    for( int i = 0; i < len; i++ ) {
        printf( "%02x,", in[i] );
    }
    printf( "\n" );
}

static bool memc( uint8_t *buf, uint8_t c, int len )
{
    for( int i = 0; i < len; i++ ) {
        if( buf[i] != c ) return true;
    }
    return false;
}

static void memrand( void *buf, int len )
{
    uint8_t *dst = buf;
    for( int i = 0; i < len; i++ ) {
        dst[i] = rand();
    }
}

static bool validate_aes( int len )
{
    char *text = "I've seen things you people wouldn't believe. Attack ships on fire off the shoulder of Orion. "
                 "I watched C-beams glitter in the dark near the Tannhauser gate. All those moments will be lost "
                 "in time, like tears in rain. Time to die.";

    srand( 42 );

    uint8_t *dst1o = malloc( len + 4096 );
    uint8_t *dst2o = malloc( len + 4096 );
    uint8_t *dst1n = malloc( len + 4096 );
    uint8_t *dst2n = malloc( len + 4096 );
    uint8_t *src   = malloc( MAX_LEN );

    assert( dst1o );
    assert( dst2o );
    assert( dst1n );
    assert( dst2n );
    assert( src );

    // Clear outputs
    memset( dst1o, 0, len + 4096 );
    memset( dst2o, 0, len + 4096 );
    memset( dst1n, 0, len + 4096 );
    memset( dst2n, 0, len + 4096 );

    // Random base iv with byte 14,15 set to zero
    uint32_t base_iv[4];
    memrand( base_iv, sizeof(base_iv) );
    ((uint8_t *)base_iv)[14] = 0;
    ((uint8_t *)base_iv)[15] = 0;

    // Random key
    AES_KEY key;
    key.rounds = 10;
    memrand( key.rd_key, sizeof(key.rd_key) );

    // Random input
    memrand( src, len );
    // Put some interesting text at start of input
    strcpy( (char *)src, text );

    // Fill output buffers with pattern
    memset( dst1o, 0xaa, len + 4096 );
    memset( dst2o, 0xaa, len + 4096 );
    memset( dst1n, 0xaa, len + 4096 );
    memset( dst2n, 0xaa, len + 4096 );

    printf( "Test len %7d: ", len );

    unsigned char ecount_buf[AES_BLOCK_SIZE];
    unsigned int num;
    uint32_t iv[4];

    // Run reference test
    memcpy( iv, base_iv, sizeof(base_iv) );
    memset( ecount_buf, 0, sizeof(ecount_buf) );
    num = 0;
    AES_ctr128_encrypt( src,   dst1o, len, &key, (uint8_t *)iv, ecount_buf, &num ); // encrypt src   -> dst1o

    memcpy( iv, base_iv, sizeof(base_iv) );
    memset( ecount_buf, 0, sizeof(ecount_buf) );
    num = 0;
    AES_ctr128_encrypt( dst1o, dst2o, len, &key, (uint8_t *)iv, ecount_buf, &num ); // decrypt dst1o -> dst2o

    // Run new code test
    srtp_crypt( (uint32_t *)src,   (uint32_t *)dst1n, len, &key, base_iv ); // encrypt src   -> dst1n
    srtp_crypt( (uint32_t *)dst1n, (uint32_t *)dst2n, len, &key, base_iv ); // decrypt dst1n -> dst2n

    bool ok = true;

    // Check for encryption/decryption errors
    if( memcmp( dst1o, dst1n, len ) ) {
        int l = len < 32 ? len : 32;
        printf( "Encrypt Fail\n" );
        hexdump( "dst1o start", dst1o,       l );
        hexdump( "dst1n start", dst1n,       l );
        hexdump( "dst1o end  ", dst1o+len-l, l );
        hexdump( "dst1n end  ", dst1n+len-l, l );
        ok = false;
    }

    if( memcmp( dst2o, dst2n, len ) ) {
        int l = len < 32 ? len : 32;
        printf( "Decrypt Fail\n" );
        hexdump( "dst2o start", dst2o,       l );
        hexdump( "dst2n start", dst2n,       l );
        hexdump( "dst2o end  ", dst2o+len-l, l );
        hexdump( "dst2n end  ", dst2n+len-l, l );
        ok = false;
    }

    // Check for buffer overruns
    if( memc( dst1o+len, 0xaa, 4096 ) ) { printf( "Encrypt overrun old\n" ); ok = false; }
    if( memc( dst1n+len, 0xaa, 4096 ) ) { printf( "Encrypt overrun new\n" ); ok = false; }
    if( memc( dst2o+len, 0xaa, 4096 ) ) { printf( "Decrypt overrun old\n" ); ok = false; }
    if( memc( dst2n+len, 0xaa, 4096 ) ) { printf( "Decrypt overrun new\n" ); ok = false; }

    if( ok ) printf( "Ok\n" );

    free( dst1o );
    free( dst2o );
    free( dst1n );
    free( dst2n );
    free( src );

    return ok;
}

static uint64_t measure_aes( int len, int cnt, bool use_ssl )
{
    uint8_t *dst = malloc( len );
    uint8_t *src = malloc( len );

    assert( dst );
    assert( src );

    memset( dst, 0xaa, len );
    memset( src, 0xaa, len );

    // Random base iv with byte 14,15 set to zero
    uint32_t base_iv[4];
    for( int i = 0; i < sizeof(base_iv)-2; i++ ) {
        ((uint8_t *)base_iv)[i] = rand()&0xff;
    }
    ((uint8_t *)base_iv)[14] = 0;
    ((uint8_t *)base_iv)[15] = 0;

    // Random key
    AES_KEY key;
    key.rounds = 10;
    for( int i = 0; i < sizeof(key.rd_key); i++ ) {
        ((uint8_t *)key.rd_key)[i] = rand()&0xff;
    }

    uint64_t t0 = clockish();

    if( use_ssl ) {

        for( int i = 0; i < cnt; i++ ) {
            unsigned char ecount_buf[AES_BLOCK_SIZE];
            unsigned int num;
            uint32_t iv[4];

            memcpy( iv, base_iv, sizeof(base_iv) );
            memset( ecount_buf, 0, sizeof(ecount_buf) );
            num = 0;
            AES_ctr128_encrypt( src, dst, len, &key, (uint8_t *)iv, ecount_buf, &num );

        }

    } else {

        for( int i = 0; i < cnt; i++ ) {

            srtp_crypt( (uint32_t *)src, (uint32_t *)dst, len, &key, base_iv );

        }

    }

    uint64_t t1 = clockish();

    return t1-t0;
}

int main( int argc, char *argv[] )
{
    int testlens[] = { 0, 1, 15, 16, 17, 4095, 4096, 4097, MAX_LEN-1, MAX_LEN };
    int testcnt = sizeof(testlens)/sizeof(testlens[0]);

    for( int i = 0; i < testcnt; i++ ) {
        bool ok = validate_aes( testlens[i] );
        if( !ok ) return 1;
    }

    for( int i = 1; i <= 64; i++ ) {
        int iters = 1000000;
        while( true ) {
            double ssl0 = measure_aes( i, iters, true );
            double xxl0 = measure_aes( i, iters, false );
            printf( "%7d %7d %.0f %.0f %f\n", i, iters, ssl0, xxl0, (ssl0-xxl0)*100/xxl0 );
            break;
        }
    }

    for( int i = 96; i <= 4096; i += 32 ) {
        int iters = 15000;
        while( true ) {
            double ssl0 = measure_aes( i, iters, true );
            double xxl0 = measure_aes( i, iters, false );
            printf( "%7d %7d %.0f %.0f %f\n", i, iters, ssl0, xxl0, (ssl0-xxl0)*100/xxl0 );
            break;
        }
    }

    for( int i = 5*1024; i <= 65536; i += 1024 ) {
        int iters = 2000;
        while( true ) {
            double ssl0 = measure_aes( i, iters, true );
            double xxl0 = measure_aes( i, iters, false );
            printf( "%7d %7d %.0f %.0f %f\n", i, iters, ssl0, xxl0, (ssl0-xxl0)*100/xxl0 );
            break;
        }
    }

    return 0;
}
